from threeML.models.tablemodel import FITSTableModel
import numpy
import collections
from threeML.models.Parameter import Parameter
import math
import scipy.integrate
import operator



class AsafModel(FITSTableModel):

    def setup(self):
        
        self.SetTableFile("/Users/jburgess/Research/3ML/threeml/models/TableModel_grid256_res8_Nr2.fits")

                
        self.functionName        = "SubPhotosphericDissipation"
        self.formula             = r"Pe'er (2005)"
        self.parameters          = collections.OrderedDict()
        self.parameters['Tau'] = Parameter('Tau',1.,1.,10.,0.1,fixed=False,nuisance=False,dataset=None)
        self.parameters['Gamma']     = Parameter('Gamma',50.,50.,500.,0.02,fixed=False,nuisance=False,dataset=None,normalization=False)
        self.parameters['LGRB']  = Parameter('LGRB',.1,100.,1e10,1,fixed=False)
        self.parameters['epsilon_d']  = Parameter('epsilon_d',.2,.2,.699999,1,fixed=False)
    
        self.ncalls              = 0
    
        
 
  
    def __call__(self,energy):
        self.ncalls             += 1
        tau                       = self.parameters['Tau'].value
        gamma                     = self.parameters['Gamma'].value
        LGRB                      = self.parameters['LGRB'].value
        ed                        = self.parameters['epsilon_d'].value
     
        return self._interpFunc((tau,gamma,LGRB,ed,energy))



from threeML.models.spectralmodel import SpectralModel

from threeML.models.Parameter import Parameter

import re, parse, parser

import collections

import numpy

class CustomSpectralModel( SpectralModel ):
    
    def setup(self, modelDef, **kwargs):
        
        if 'function' in kwargs.keys():
            
            self.userFunction = kwargs['function']
        
        else:
            
            self.userFunction = None
        
        #Divide in lines
        lines = modelDef.replace("\r","\n").split("\n")
        
        #Remove empty lines
        
        filtered = filter(lambda x: not re.match(r'^\s*$', x), lines)
        
        if len(filtered) <= 2:
            
            raise RuntimeError("You have to provide at least a name:formula line and a init line.")
        
        #Parse first line: model name and formula
        
        r = parse.parse('{name}:{formula}', filtered[0].replace(" ",""))
        
        if r is None:
            
            raise RuntimeError("Could not understand first line: %s" %filtered[0])
        
        self.functionName = r['name']
        self.formula = r['formula']
        
        #Second line: init values
        
        r = parse.parse('init:{initValues}', filtered[1].replace(" ",""))
        
        if r is None:
            
            raise RuntimeError("Could not understand second line: %s" %filtered[1])
        
        paramNames = []
        initValues = []
        
        for pp in r['initValues'].split(","):
            
            try:
            
                name,value = pp.split("=")
            
            except:
                
                raise RuntimeError("Cannot understand init value %s" % pp)
            
            if name=='e' or name=='E':
                
                raise RuntimeError("You cannot use 'e' or 'E' as parameter names. They indicate energy.")
            
            paramNames.append( name )
            initValues.append( float(value) )
        
        #Make the parameters dictionary
        
        self.parameters = collections.OrderedDict()
        
        for p,v in zip( paramNames, initValues ):
            
            #If the name is all upper case, it is a scale or normalization
            #parameter
            if p.upper()==p:
                
                norm = True
            
            else:
                
                norm = False
            
            self.parameters[p] = Parameter(p, v, None, None, v / 10.0, normalization=norm)
        
        #Third line: boundaries
        
        if len(filtered) > 2:
        
            r = parse.parse('bounds:{boundsSpec}', filtered[2].replace(" ",""))
            
            if r is None:
                
                raise RuntimeError("Could not understand third line: %s" %(filtered[2]))
            
            boundsExpr = r['boundsSpec'].split(",")
            
            for p in self.parameters.keys():
                
                #Find out if there is a bound specification for this parameter
                thisExpr = filter(lambda x:x.find(p) >= 0, boundsExpr)
                
                if len(thisExpr)==0:
                    
                    #No, just continue
                    continue
                
                elif len(thisExpr) > 1:
                    
                    raise RuntimeError("More than one bounds specification for parameter %s" % p)
                
                #If we get here, there is a specification for the bounds
                #Parse it
                pexpr = thisExpr[0]
                
                #Remove it from the boundsExpr list
                boundsExpr.pop( boundsExpr.index(pexpr) )
                
                r = parse.parse('{minvalue}<={pname}<={maxvalue}', pexpr)
            
                if r is not None:
                
                    #Max and min specified
                    minValue = float( r['minvalue'] )
                    maxValue = float( r['maxvalue'] )
                
                else:
                    
                    r = parse.parse('{pname}<={maxvalue}',pexpr)
                
                    if r is not None:
                        
                        minValue = None
                        maxValue = float( r['maxvalue'] )
                    
                    else:
                        
                        r = parse.parse('{pname}>={minvalue}', pexpr)
                        
                        if r is not None:
                            
                            minValue = float( r['minvalue'] )
                            maxValue = None
                        
                        else:
                            
                            raise RuntimeError("Could not understand expression %s" %pexpr)
                
                self.parameters[p].setBounds(minValue,maxValue)
            
            #If the boundsExpr is not empty at this point, it means that there is a bounds
            #specification for a parameter which does not exists...
            if len(boundsExpr) > 0:
                
                raise RuntimeError("One or more bounds refer to non-existent parameters.")
            
        pass #End of boundaries parsing
        
        #Now sanitize the formula removing spaces
        
        self.formula = "".join( self.formula.split() )
        
        self.finalSetup()
        
    
    def __getstate__( self ):
        
        #This is used by pickle to get the state of the current
        #instance. The return value of this method will be used
        #by the __setstate__ method on the other side (the remote
        #side) to recreate the class
        
        d = {}
        
        d['formula'] = self.formula
        d['userFunction'] = self.userFunction
        d['parameters'] = self.parameters
        
        return d
    
    def __setstate__( self, state ):
        
        self.formula = state['formula']
        self.userFunction = state['userFunction']
        self.parameters = state['parameters']
        
        self.finalSetup()
        
    def finalSetup( self ):
        
        #Now compile the formula
        
        self._compiledFormula = parser.expr(self.formula).compile()
        
        #Create a local dictionary to store variables which will be used
        #by the formula just compiled
        self._locals = {}
        
        #Create pointers to normal numpy functions, so that they can be used
        #in the formula
        self._locals['log'] = numpy.log
        self._locals['log10'] = numpy.log10
        self._locals['power'] = numpy.power
        self._locals['exp'] = numpy.exp
        
        #If the user specified a user-defined formula, add it to the
        #dictionary
        
        if not self.userFunction is None:
            
            fname = self.userFunction.__name__
            
            self._locals[fname] = self.userFunction
        
        #Test it with default parameters
        
        e = numpy.linspace(1.0,100,10)
        
        try:
            
            testValue = self.__call__( e )
        
        except:
                        
            print("\n\nTest did not work. Check your syntax (Remember: parameters are case-sensitive!).\n\n")
            
            raise
        
        else:
            
            if numpy.sum( ~numpy.isfinite(testValue) )==0:
                
                #Everything seems to be fine
                
                pass
            
            else:
                
                raise RuntimeError("The compiled function returned NaN or inf. Check your syntax.")
    
    def __call__(self, e):
                    
        for p in self.parameters.keys():
                
            self._locals[p] = self.parameters[p].value
            
        self._locals['e'] = numpy.array( e )
            
        return eval( self._compiledFormula, self._locals )



#author H. Zhou (hzhou1@mtu.edu)

from threeML.models.spatialmodel import SpatialModel
from threeML.models.Parameter import Parameter, SpatialParameter
import numpy as np
import scipy.stats
from angsep import angsep

import collections


class Diffusion(SpatialModel):
    def setup(self):
        self.functionName        = "Diffusion"
        self.formula             = r'$f({\rm RA, Dec}) = \left(\frac{180^\circ}{\pi}\right)^2 \frac{1.2154}{\sqrt{\pi^3} \rdiff ({\rm angsep} ({\rm RA, Dec, RA_0, Dec_0})+0.06 \rdiff)} \, {\rm exp}\left(-\frac{{\rm angsep}^2 ({\rm RA, Dec, RA_0, Dec_0})}{\rdiff^2} \right) $'
        self.parameters          = collections.OrderedDict()
        self.parameters['RA0']     = Parameter('RA0',1.,0,360,0.1,fixed=False,nuisance=False,dataset=None)
        self.parameters['Dec0']     = Parameter('Dec0',1.,-90,90,0.1,fixed=False,nuisance=False,dataset=None)
        self.parameters['rdiff'] = SpatialParameter('rdiff',1.,0,20,0.01,fixed=False,nuisance=False,dataset=None)
        self.parameters['delta'] = SpatialParameter('delta',0.5,0.3,0.6,0.01,fixed=True,nuisance=False,dataset=None)
            
        self.ncalls              = 0

  
    def __call__(self,RA,Dec,energy):
        self.ncalls             += 1
        RA0                         = self.parameters['RA0'].value
        Dec0                        = self.parameters['Dec0'].value
        #energy in kev -> TeV
        rdiff                       = self.parameters['rdiff'].value * np.power(np.divide(energy,2e10),(self.parameters['delta'].value - 1.)/2.*(0.54+0.046*np.log10(np.divide(energy,1e9))))
        
        return np.maximum( np.power(180/np.pi,2)*1.2154/(np.pi * np.sqrt(np.pi) * rdiff * (angsep(RA,Dec,RA0,Dec0) + 0.06 * rdiff)) * np.exp(-1. * np.power(angsep(RA,Dec,RA0,Dec0),2)/rdiff**2), 1e-30)

    def integratedFlux(self,energy):
    
        return 1.
    
    def getBoundaries(self):
        
        #Truncate the function at the max of rdiff allowed
        
        maxRdiff = self.parameters['rdiff'].maxValue
        
        minDec = max(-90.,self.parameters['Dec0'].value - maxRdiff)
        maxDec = min(90.,self.parameters['Dec0'].value + maxRdiff)

        maxAbsDec = max(np.absolute(minDec),np.absolute(maxDec))
        if maxAbsDec > 89. or maxRdiff/np.cos(maxAbsDec*np.pi/180.) >= 180.:
            minRa = 0.
            maxRa = 360.
        else:
            minRa = self.parameters['RA0'].value - maxRdiff/np.cos(maxAbsDec*np.pi/180.)
            maxRa = self.parameters['RA0'].value + maxRdiff/np.cos(maxAbsDec*np.pi/180.)
            if minRa < 0.:
                minRa = minRa + 360.
            elif maxRa > 360.:
                maxRa = maxRa - 360.
        
        return minRa, maxRa, minDec, maxDec
                      



        




import astropy.io.fits as pyfits
import xml.etree.ElementTree as ET
import os

from threeML.plugins.gammaln import logfactorial
from threeML.io import fileUtils
from threeML.plugins.GenericOGIPLike import GenericOGIPLike

import numpy
from threeML.plugins.ogip import OGIPPHA
from threeML.plugin_prototype import PluginPrototype
from threeML.models.Parameter import Parameter
from threeML.minimizer import minimization
import scipy.integrate

import warnings
import collections

__instrument_name = "Swift XRT"

class SwiftXRTLike( GenericOGIPLike ):
  
  def __init__(self, name, phafile, bkgfile, rspfile, arffile):
      
      super(SwiftXRTLike, self).__init__(name, phafile, bkgfile, rspfile, arffile)



from threeML.plugin_prototype import PluginPrototype

from threeML.models.Parameter import Parameter

from threeML.io.fileUtils import fileExistingAndReadable, sanitizeFilename

from cthreeML.pyModelInterface import pyToCppModelInterface

from hawc import liff_3ML

import os, sys, collections

import matplotlib.pyplot as plt
from matplotlib import gridspec

import numpy

defaultMinChannel = 0
defaultMaxChannel = 9

__instrument_name = "HAWC"

class HAWCLike( PluginPrototype ):
    
    def __init__( self, name, maptree, response, ntransits = None, **kwargs ):
        
        #This controls if the likeHAWC class should load the entire
        #map or just a small disc around a source (faster).
        #Default is the latter, which is way faster. LIFF will decide
        #autonomously which ROI to use depending on the source model
        
        self.fullsky = False
        
        if 'fullsky' in kwargs.keys():
            
            self.fullsky = bool( kwargs['fullsky'] )
        
        self.name = str( name )
        
        #Sanitize files in input (expand variables and so on)
        
        self.maptree = os.path.abspath( sanitizeFilename( maptree ) )
        
        self.response = os.path.abspath( sanitizeFilename( response ) )

        #
        self.ntransits = ntransits

        #Check that they exists and can be read
        
        if not fileExistingAndReadable( self.maptree ):
        
            raise IOError("MapTree %s does not exist or is not readable" % maptree)
    
        if not fileExistingAndReadable( self.response ):
        
            raise IOError("Response %s does not exist or is not readable" % response)
        
        #Post-pone the creation of the LIFF instance to when
        #we have the likelihood model
        
        self.instanced = False
        
        #Default value for minChannel and maxChannel
        
        self.minChannel = int( defaultMinChannel )
        self.maxChannel = int( defaultMaxChannel )
        
        #By default the fit of the CommonNorm is deactivated
        
        self.deactivateCommonNorm()
        
        #This is to keep track of whether the user defined a ROI or not
        
        self.roi_ra = None
        
        #Further setup
        
        self.__setup()

    def setROI(self, ra, dec, radius, fixedROI=False):
        
        self.roi_ra = ra
        self.roi_dec = dec
        
        self.roi_radius = radius

        self.fixedROI = fixedROI
    
    def __setup(self):
        
        #I put this here so I can use it both from the __init__ both from
        #the __setstate__ methods
        
        #Create the dictionary of nuisance parameters
        
        self.nuisanceParameters = collections.OrderedDict()
        self.nuisanceParameters['CommonNorm'] = Parameter("CommonNorm",1.0,0.5,1.5,0.01,
                                                           fixed=True,nuisance=True)
    
    def __getstate__(self):
        
        #This method is used by pickle before attempting to pickle the class
        
        #Return only the objects needed to recreate the class
        #IN particular, we do NOT return the theLikeHAWC class,
        #which is not pickeable. It will instead be recreated
        #on the other side
        
        d = {}
        
        d['name']= self.name
        d['maptree'] = self.maptree
        d['response'] = self.response
        d['ntransits'] = self.ntransits
        d['model'] = self.model
        d['minChannel'] = self.minChannel
        d['maxChannel'] = self.maxChannel
        
        d['roi_ra'] = self.roi_ra
        
        if self.roi_ra is not None:
        
            d['roi_dec'] = self.roi_dec
            d['roi_radius'] = self.roi_radius
        
        return d
    
    def __setstate__( self, state ):
        
        #This is used by pickle to recreate the class on the remote
        #side
        name = state['name']
        maptree = state['maptree']
        response = state['response']
        ntransits = state['ntransits']
        
        #Now report the class to its state
        
        self.__init__( name, maptree, response, ntransits )
        
        if state['roi_ra'] is not None:
        
            self.setROI( state['roi_ra'], state['roi_dec'], state['roi_radius'], state['fixedROI'] )
        
        self.setActiveMeasurements( state['minChannel'], state['maxChannel'] )
        
        self.setModel( state['model'] )

    
    def setActiveMeasurements( self, minChannel, maxChannel ):
        
        self.minChannel = int( minChannel )
        self.maxChannel = int( maxChannel )
        
        if self.instanced:
            
            sys.stderr.write("Since the plugins was already used before, the change in active measurements" +
                             "will not be effective until you create a new JointLikelihood or Bayesian" +
                             "instance")
        
    def setModel( self, LikelihoodModelInstance ):
        '''
        Set the model to be used in the joint minimization. Must be a LikelihoodModel instance.
        '''
        
        #Instance the python - C++ bridge
        self.model = LikelihoodModelInstance
        
        self.pymodel = pyToCppModelInterface( self.model )
        
        #Now init the HAWC LIFF software
    
        try:
            
            #Load all sky
            #(ROI will be defined later)
            
            if self.ntransits is None:
                self.theLikeHAWC = liff_3ML.LikeHAWC(self.maptree,
                                                     self.response,
                                                     self.pymodel,
                                                     self.minChannel,
                                                     self.maxChannel,
                                                     self.fullsky)
            
            else:
                self.theLikeHAWC = liff_3ML.LikeHAWC(self.maptree,
                                                     self.ntransits,
                                                     self.response,
                                                     self.pymodel,
                                                     self.minChannel,
                                                     self.maxChannel,
                                                     self.fullsky)
            
            if self.roi_ra is None and self.fullsky:
                
                raise RuntimeError("You have to define a ROI with the setROI method")
            
            if self.roi_ra is not None and self.fullsky:
            
                self.theLikeHAWC.SetROI( self.roi_ra, self.roi_dec, self.roi_radius, self.fixedROI )
            
        except:
            
            print("Could not instance the LikeHAWC class from LIFF. " +
                               "Check that HAWC software is working")
            
            raise
        
        else:
            
            self.instanced = True
                    
        #Now set a callback in the CommonNorm parameter, so that if the user or the fit
        #engine or the Bayesian sampler change the CommonNorm value, the change will be
        #propagated to the LikeHAWC instance
        
        self.nuisanceParameters['CommonNorm'].setCallback( self._CommonNormCallback )
        
    
    def _CommonNormCallback( self ):
        
        self.theLikeHAWC.SetCommonNorm( self.nuisanceParameters['CommonNorm'].value )
    
    def getName(self):
        '''
        Return a name for this dataset (likely set during the constructor)
        '''
        return self.name
    
    def activateCommonNorm( self ):
        
        self.fitCommonNorm = True
    
    def deactivateCommonNorm( self ):
        
        self.fitCommonNorm = False
    
    def getLogLike(self):
        
        '''
        Return the value of the log-likelihood with the current values for the
        parameters
        '''
        
        self.pymodel.update()
        
        logL = self.theLikeHAWC.getLogLike( self.fitCommonNorm )
                
        return logL

    def calcTS(self):
        
        '''
        Return the value of the log-likelihood test statistic, defined as
        2*[log(LL_model) - log(LL_bkg)]
        '''
        
        self.pymodel.update()
        
        TS = self.theLikeHAWC.calcTS( self.fitCommonNorm )
                
        return TS
  
    def getNuisanceParameters(self):
        '''
        Return a list of nuisance parameters. Return an empty list if there
        are no nuisance parameters
        '''
        
        return self.nuisanceParameters.keys()
  
    def innerFit(self):
        
        self.theLikeHAWC.SetBackgroundNormFree( self.fitCommonNorm )
        
        self.pymodel.update()
        
        logL = self.theLikeHAWC.getLogLike( self.fitCommonNorm )
        
        self.nuisanceParameters['CommonNorm'].setValue( self.theLikeHAWC.CommonNorm() )
        
        return logL
    
    def display( self, radius = 2.0 ):
        
        figs = []
        
        nsrc = self.model.getNumberOfPointSources()
        
        for srcid in range(nsrc):
            
            ra, dec = self.model.getPointSourcePosition( srcid )
            
            model = numpy.array( self.theLikeHAWC.GetTopHatExpectedExcesses( ra, dec, radius ) )
            
            signal = numpy.array(self.theLikeHAWC.GetTopHatExcesses( ra, dec, radius ) )
            
            bkg = numpy.array( self.theLikeHAWC.GetTopHatBackgrounds( ra, dec, radius ) )
            
            total = signal + bkg
            
            fig = plt.figure()
            
            gs = gridspec.GridSpec(2,1, height_ratios=[2,1])
            gs.update(hspace=0)
            
            sub = plt.subplot( gs[0] )
            
            nHitBins = numpy.arange( self.minChannel, self.maxChannel + 1 )
            
            sub.errorbar( nHitBins, total, yerr=numpy.sqrt(total), 
                          capsize=0, color='black', label='Observation',
                          fmt='.')

            sub.plot( nHitBins, model + bkg, label='Model + bkg')
            
            plt.legend( bbox_to_anchor=(1.05, 1), loc=2, numpoints=1)
            
            #Residuals
            
            sub1 = plt.subplot( gs[1] )
            
            #Using model variance to account for low statistic
            
            resid = ( signal - model ) / model
            
            sub1.axhline( 0, linestyle='--' )
            
            
            sub1.errorbar( nHitBins, resid,
                           yerr=numpy.sqrt(total) / model,
                           capsize=0, fmt='.' )
            
            
            sub.set_xlim( [nHitBins.min() - 0.5, nHitBins.max() + 0.5] )
            
            sub.set_yscale("log", nonposy='clip')
            
            sub.set_ylabel("Counts per bin") 
            
            #sub1.set_xscale("log")
            
            sub1.set_xlabel("Analysis bin")
            
            sub1.set_ylabel(r"$\frac{excess - mod.}{mod.}$",fontsize=20)
            
            sub1.set_xlim( [nHitBins.min() - 0.5, nHitBins.max() + 0.5] )
            
            sub.set_xticks( [] )
            sub1.set_xticks( nHitBins )
            
            figs.append(fig)
            
        return figs
    
    def writeModelMap(self, fileName, poisson=False):
        
        self.theLikeHAWC.WriteModelMap( fileName, poisson )
    
    def writeResidualMap(self, fileName):
        
        self.theLikeHAWC.WriteResidualMap( fileName )
  




import os

def fileExistingAndReadable( filename ):
    
    if os.path.exists( filename ):
        
	#Try to open it
	
	try:
	    
	    with open( filename ):
	    
	        pass
	
	except:
	    
	    return False
	
	else:
	
	    return True

    else:
        
	return False


def sanitizeFilename( filename ):
    
    return os.path.expandvars( os.path.expanduser( filename ) )



#!/usr/bin/env python

import os
import sys

from setuptools import setup, Extension

# Get the version number
execfile('threeML/version.py')

# Now a global __version__ is available

# This list will contain the messages to print just before the end of the setup
# so that the user actually note them, instead of loosing them in the tons of
# messages of the build process

final_messages = ["REMEMBER: if you want to use C/C++ plugins (HAWC) you have to install also cthreeML"]

setup(

    name="threeML",

    packages=['threeML',
              'threeML/exceptions',
              'threeML/bayesian',
              'threeML/minimizer',
              'threeML/models',
              'threeML/models/fluxModels',
              'threeML/models/spatialModels',
              'threeML/plugins',
              'threeML/classicMLE',
              'threeML/catalogs',
              'threeML/io',
              'threeML/utils',
              'threeML/parallel',
              'threeML/config'],

    version=__version__,

    description="The Multi-Mission Maximum Likelihood framework",

    long_description="3ML can be used for single or multi-instrument likelihood modeling or Bayesian inference",
    
    license='BSD-3',

    author='Giacomo Vianello',

    author_email='giacomo.vianello@gmail.com',

    url='https://github.com/giacomov/3ML',

    download_url='https://github.com/giacomov/3ML/archive/%s' % __version__,

    keywords=['Likelihood', 'Multi-mission', '3ML', 'HAWC', 'Fermi', 'HESS', 'joint', 'fit', 'bayesian',
              'multi-wavelength'],

    classifiers=[],

    # Install configuration file in user home and in the package repository

    data_files=[(os.path.join(os.path.expanduser('~'), '.threeML'), ["threeML/config/threeML_config.yml"]),
                ('threeML/config', ["threeML/config/threeML_config.yml"])
                ],

    install_requires=[
        'numpy >= 1.6',
        'scipy',
        'emcee',
        'astropy>=1.0.3',
        'matplotlib',
        'ipython',
        'uncertainties',
        'pyyaml',
        'dill',
        'parse',
        'iminuit'
    ])

# Now print the final messages if there are any

if len(final_messages) > 0:
    print("\n#############")
    print("FINAL NOTES:")
    print("#############")

    print("\n".join(final_messages))



from __future__ import print_function

import sys, time

class ProgressBar:
    def __init__(self, iterations):
        self.iterations = iterations
        self.prog_bar = '[]'
        self.fill_char = '*'
        self.width = 50
        self.startTime = time.time()
        self.lastIter = 0
        self.__update_amount(0)

    def animate(self, iter):
        print('\r', self, end='')
        sys.stdout.flush()
        self.lastIter = iter
        self.update_iteration(iter + 1)
    
    def increase(self):
        
        self.animate(self.lastIter + 1)
    
    def update_iteration(self, elapsed_iter):
        elapsed_iter = min(elapsed_iter, self.iterations)
        self.__update_amount((elapsed_iter / float(self.iterations)) * 100.0)
        self.prog_bar += '  %d of %s completed in %.1f s' % (elapsed_iter, self.iterations, time.time() - self.startTime)
        #sys.stdout.flush()

    def __update_amount(self, new_amount):
        percent_done = min(int(round((new_amount / 100.0) * 100.0)), 100)
        all_full = self.width - 2
        num_hashes = int(round((percent_done / 100.0) * all_full))
        self.prog_bar = '[' + self.fill_char * num_hashes + ' ' * (all_full - num_hashes) + ']'
        pct_place = (len(self.prog_bar) // 2) - len(str(percent_done))
        pct_string = '%d%%' % percent_done
        self.prog_bar = self.prog_bar[0:pct_place] + \
            (pct_string + self.prog_bar[pct_place + len(pct_string):])

    def __str__(self):
        return str(self.prog_bar)



#author L. Tibaldo (ltibaldo@slac.stanford.edu)

from threeML.models.spatialmodel import SpatialModel
from threeML.models.Parameter import Parameter, SpatialParameter
import numpy as np
import scipy.stats
from angsep import angsep

import collections


class SimpleGaussian(SpatialModel):

    def setup(self):
        self.functionName        = "Gaussian"
        self.formula             = r'$f({\rm RA, Dec}) = \left(\frac{180^\circ}{\pi}\right)^2 \frac{1}{2\pi \sigma^2} \, {\rm exp}\left(-\frac{{\rm angsep}^2 ({\rm RA, Dec, RA_0, Dec_0})}{2\sigma^2} \right) $'
        self.parameters          = collections.OrderedDict()
        self.parameters['RA0']     = Parameter('RA0',1.,0,360,0.1,fixed=False,nuisance=False,dataset=None)
        self.parameters['Dec0']     = Parameter('Dec0',1.,-90,90,0.1,fixed=False,nuisance=False,dataset=None)
        self.parameters['sigma'] = SpatialParameter('sigma',0.1,0,20,0.01,fixed=False,nuisance=False,dataset=None)
            
        self.ncalls              = 0

  
    def __call__(self,RA,Dec,energy):
        self.ncalls             += 1
        RA0                         = self.parameters['RA0'].value
        Dec0                        = self.parameters['Dec0'].value
        sigma                       = self.parameters['sigma'].getValue(energy)
        
        return np.maximum( np.power(180/np.pi,2)*1./(2 * np.pi * sigma**2) * np.exp(-0.5 * np.power(angsep(RA,Dec,RA0,Dec0),2)/sigma**2), 1e-30)

    def integratedFlux(self,energy):
    
        return 1.
    
    def getBoundaries(self):
        
        #Truncate the gaussian at 2 times the max of sigma allowed
        
        maxSigma = self.parameters['sigma'].maxValue
        
        minDec = max(-90.,self.parameters['Dec0'].value - 2 * maxSigma)
        maxDec = min(90.,self.parameters['Dec0'].value + 2 * maxSigma)

        maxAbsDec = max(np.absolute(minDec),np.absolute(maxDec))
        if maxAbsDec > 89. or 2*maxSigma/np.cos(maxAbsDec*np.pi/180.) >= 180.:
            minRa = 0.
            maxRa = 360.
        else:
            minRa = self.parameters['RA0'].value - 2*maxSigma/np.cos(maxAbsDec*np.pi/180.)
            maxRa = self.parameters['RA0'].value + 2*maxSigma/np.cos(maxAbsDec*np.pi/180.)
            if minRa < 0.:
                minRa = minRa + 360.
            elif maxRa > 360.:
                maxRa = maxRa - 360.
        
        return minRa, maxRa, minDec, maxDec
                      


class MultiVariateGaussian(SpatialModel):
    
    def setup(self):
        self.functionName        = "Multivariate Gaussian"
        self.formula             = r'$f(\vec{x}) = \left(\frac{180^\circ}{\pi}\right)^2 \frac{1}{2\pi \sqrt{\det{\Sigma}}} \, {\rm exp}\left( -\frac{1}{2} (\vec{x}-\vec{x}_0)^\intercal \cdot \Sigma^{-1}\cdot (\vec{x}-\vec{x}_0)\right) \\ \vec{x}_0 = ({\rm RA}_0,{\rm Dec}_0)\\ \Lambda = \left( \begin{array}{cc} \sigma^2 & 0 \\ 0 & \sigma^2 (1-e^2) \end{array}\right) \\ U = \left( \begin{array}{cc} \cos \theta & -\sin \theta \\ \sin \theta & cos \theta \end{array}\right) \\\Sigma = U\Lambda U^\intercal$'
        self.parameters          = collections.OrderedDict()
        self.parameters['RA0']     = Parameter('RA0',1.,0,360,0.1,fixed=False,nuisance=False,dataset=None)
        self.parameters['Dec0']     = Parameter('Dec0',1.,-90,90,0.1,fixed=False,nuisance=False,dataset=None)
        self.parameters['sigma'] = SpatialParameter('sigma',0.1,0,10,0.01,fixed=False,nuisance=False,dataset=None)
        self.parameters['eccentricity'] = SpatialParameter('eccentricity',0.7,0,1,0.01,fixed=False,nuisance=False,dataset=None)
        self.parameters['angle'] = SpatialParameter('angle',0.,0,180,1.,fixed=False,nuisance=False,dataset=None)
        
        
        self.ncalls              = 0
    
    
    def __call__(self,RA,Dec,energy):
        self.ncalls             += 1
        RA0                         = self.parameters['RA0'].value
        Dec0                        = self.parameters['Dec0'].value
        sigma                       = self.parameters['sigma'].getValue(energy)
        eccentricity                = self.parameters['eccentricity'].getValue(energy)
        angle                       = np.deg2rad(self.parameters['angle'].getValue(energy))
        
        
        sigma1=sigma**2
        sigma2=sigma1*(1-eccentricity**2)
        rot=np.array([[np.cos(angle),-np.sin(angle)],[np.sin(angle),np.cos(angle)]])
        cov=np.dot(rot,np.dot(np.array([[sigma1,0],[0,sigma2]]),rot.T))
        return np.maximum(np.power(180/np.pi,2)*scipy.stats.multivariate_normal.pdf(np.array([Dec,RA]).T, mean=np.array([RA0,Dec0]), cov=cov), 1e-30)
   
    def integratedFlux(self,energy):
                                                 
        return 1.



        




from threeML.bayesian import Priors

import numpy

import scipy.stats

class Parameter(object):
  def __init__(self,name,initValue,minValue,maxValue,delta,**kwargs):
    self.name                 = str(name)
    self.value                = initValue
    self.minValue             = minValue
    self.maxValue             = maxValue
    self.delta                = delta
    self.unit                 = ''
    
    self.fixed                = False
    self.nuisance             = False
    self.dataset              = None
    self.callback             = []
    self.normalization        = False
    
    for k,v in kwargs.iteritems():
      if(k.lower()=='unit'):
        self.unit             = str(v)
      elif(k.lower()=='normalization'):
        self.normalization    = bool(v)
      elif(k.lower()=='fixed'):
        self.fixed            = bool(v)
      elif(k.lower()=='nuisance'):
        self.nuisance         = bool(v)
      elif(k.lower()=='dataset'):
        self.dataset          = v
      pass
    pass
    
    #Default prior is a uniform prior
    if(self.normalization):
      #This is a scale parameter
      self.setPrior(Priors.LogUniformPrior(self.minValue,self.maxValue))
    else:
      self.setPrior(Priors.UniformPrior(self.minValue,self.maxValue))
  pass
  
  def setCallback(self,callback):
    #The callback functions will be executed on any parameter value change
    self.callback.append(callback)
  pass

  def __eq__(self,value):
    self.setValue(value)
    
  def __repr__(self):
    if(self.fixed):
      ff                      = "fixed"
    else:
      ff                      = "free"
    pass
    
    return "%20s: %10g %10g %10g %10g %s %s" %(self.name,self.value,self.minValue,self.maxValue,self.delta,ff,self.unit)
  pass
  
  def getRandomizedValue( self, var = 0.1 ):
    
    #Get a value close to the current value, but not identical
    #(used for the inizialization of Bayesian samplers)
    
    if ( self.minValue is not None ) or ( self.maxValue is not None ):
        
        #Bounded parameter. Use a truncated normal so we are guaranteed
        #to have a random value within the boundaries
        
        std = abs( var * self.value )
        
        if self.minValue is not None:
            
            a = ( self.minValue - self.value ) / std 
        
        else:
            
            a = - numpy.inf
        
        if self.maxValue is not None:
            
            b = ( self.maxValue - self.value ) / std 
        
        else:
            
            b = numpy.inf
              
        sample = scipy.stats.truncnorm.rvs( a, b, loc = self.value, scale = std, size = 1)
        
        if sample < self.minValue or sample > self.maxValue:
            
            raise RuntimeError("BUG!!")
        
        return sample[0]
    
    else:
        
        #The parameter has no boundaries
        
        return numpy.random.normal( self.value, var * self.value )
  
  def getValue(self):
    
    return self.value
  
  def getPriorValue( self ):
    
    return self.prior( self.value )
  
  def setValue(self,value):
    self.value                = float(value)
    
    if(abs(self.delta) > 0.2*abs(self.value)):
      #Fix the delta to be less than 50% of the value
      self.delta              = 0.2 * self.value
    
    for c in self.callback:
      c()
  pass
  
  def setBounds(self,minValue,maxValue):
    self.minValue             = minValue
    self.maxValue             = maxValue
    self.prior.setBounds(minValue,maxValue)
  pass
  
  def setDelta(self,delta):
    self.delta                = delta
  pass
  
  def setPrior(self,prior):
    self.prior                = prior
  pass
  
  def setDataset(self,dataset):
    self.dataset              = dataset
  
  def fix(self):
    self.fixed                = True
  pass
  
  def free(self):
    self.fixed                = False
  pass
  
  def isNuisance(self):
    return self.nuisance
  
  def isNormalization(self):
    return self.normalization
  
  def isFixed(self):
    return self.fixed
  
  def isFree(self):
    return (not self.fixed)
pass


class SpatialParameter(object):
    
    #this class provides a place holder for spatial parameters that vary with energy, for the moment works exactly as the regular parameter with value indepedent from energy
    def __init__(self,name,initValue,minValue,maxValue,delta,**kwargs):
        self.name                 = str(name)
        self.value                = initValue
        self.minValue             = minValue
        self.maxValue             = maxValue
        self.delta                = delta
        self.unit                 = ''
        
        self.fixed                = False
        self.nuisance             = False
        self.dataset              = None
        self.callback             = []
        self.normalization        = False
        
        for k,v in kwargs.iteritems():
            if(k.lower()=='unit'):
                self.unit             = str(v)
            elif(k.lower()=='normalization'):
                self.normalization    = bool(v)
            elif(k.lower()=='fixed'):
                self.fixed            = bool(v)
            elif(k.lower()=='nuisance'):
                self.nuisance         = bool(v)
            elif(k.lower()=='dataset'):
                self.dataset          = v
            pass
        pass

        #Default prior is a uniform prior
        if(self.normalization):
        #This is a scale parameter
            self.setPrior(Priors.LogUniformPrior(self.minValue,self.maxValue))
        else:
            self.setPrior(Priors.UniformPrior(self.minValue,self.maxValue))
    pass
                
    def setCallback(self,callback):
        #The callback functions will be executed on any parameter value change
        self.callback.append(callback)
    pass
        
    def __eq__(self,value):
        self.setValue(value)
    
    def __repr__(self):
        if(self.fixed):
            ff                      = "fixed"
        else:
            ff                      = "free"
        pass
    
        return "%20s: %10g %10g %10g %10g %s %s" %(self.name,self.value,self.minValue,self.maxValue,self.delta,ff,self.unit)
        pass
            
    def setValue(self,value):
        self.value                = float(value)
    
        if(abs(self.delta) > 0.2*abs(self.value)):
        #Fix the delta to be less than 50% of the value
            self.delta              = 0.2 * self.value

        for c in self.callback:
            c()
    pass
    
    def getValue(self,energy):
        return self.value
    pass
                
    def setBounds(self,minValue,maxValue):
        self.minValue             = minValue
        self.maxValue             = maxValue
        self.prior.setBounds(minValue,maxValue)
    pass
            
    def setDelta(self,delta):
        self.delta                = delta
    pass
        
    def setPrior(self,prior):
        self.prior                = prior
    pass
        
    def setDataset(self,dataset):
        self.dataset              = dataset
    
    def fix(self):
        self.fixed                = True
    pass
            
    def free(self):
        self.fixed                = False
    pass
        
    def isNuisance(self):
        return self.nuisance
    
    def isNormalization(self):
        return self.normalization
        
    def isFixed(self):
        return self.fixed
    
    def isFree(self):
        return (not self.fixed)
pass




import numpy
import collections
from Parameter import Parameter
#import fancyDisplay
from IPython.display import display, Latex, HTML
import math
import scipy.integrate
import operator
#import numexpr
import abc
import matplotlib.pyplot as plt

from threeML.io.Table import Table


class ModelValidate(object):

    def __init__(self,model):

        self.keVtoErg                 = 1.60217657E-9
        self.checks = ["'self.functionName' not set in *setup* ",\
                    "'self.formula' not set in *setup*",\
                    "'self.parameters' not set in *setup*"\
                    ]
        checks2 = "'self.parameters' is NOT of type collections.OrderedDict"

        self.tt = [hasattr(model,'functionName'),\
                   hasattr(model,'formula'),\
                   hasattr(model,'parameters')]


        if self.tt[2]: #Check that parameters exist
            
            self.checks.append(checks2)
            self.tt.append(isinstance(model.parameters,collections.OrderedDict))

        else:

            self._printSetupErr()
            return

        if self.tt[3]: #Check that the parameters are right type

            for p in model.parameters.keys():

                self.checks.append("self.parameters[%s] is not an instance Parameter"%p)
                self.tt.append(isinstance(model.parameters[p],Parameter))
                
        self._printSetupErr()



    
    def _printSetupErr(self):

        # negate the things which are correct to return the errors
        self.tt = ~numpy.array(self.tt)

        #Turn the checks into a numpy array
        self.checks =numpy.array(self.checks)

        #Now select only the problems found via the index table
        problems = self.checks[self.tt]

        if len(problems) == 0:
            return

        print "Error: The SpectralModel class is not proper!"
        print "The following problems were found:"
        print
        for prob in problems:
                print "\t"+prob

        raise RuntimeError("Correct the SpectralModel definition!")        
          
        



class SpectralModel(object):
    __metaclass__           = abc.ABCMeta # Make the user code certain
    

    def __init__(self, *args, **kwargs):

        import collections        
        
        self.setup(*args, **kwargs)


        self._validate() # Raise runtime error if the users

        for k,v in self.parameters.iteritems():

            object.__setattr__(self,k,v)



        
    @abc.abstractmethod
    def setup(*args,**kwargs):
        # virtual member implemented by the user
        pass
    
    def _validate(self):
        '''
        Member called to check that the model has all the proper
        attributes after the setup is called in the __init__ member
        '''
    
        val = ModelValidate(self)
    
    def __getattr__(self, item):
     
        
        
        if self.__dict__.has_key("parameters"):
            if self.parameters.has_key(item):
                return self.parameters[item]
            else:
                object.__getattribute__(self,item)
        else:
            object.__getattribute__(self,item)
        
    def __setattr__(self,item,value):

                
        
        if self.__dict__.has_key("parameters"):
            
            if self.parameters.has_key(item):
                (self.parameters[item]).setValue(value)
            else:
                object.__setattr__(self,item,value)
                
        else:
            
            object.__setattr__(self,item,value)

    # Removed by J. Michael to switch to a different scheme
    #def __getitem__(self,argument):
    #    return self.parameters[argument]


    def display(self,emin=10.,emax=5000.,logscale=True,fluxType="vfv",**kwargs):
        '''
        Display the model 

        '''

        eIndxDic = {"ph":0.,"ene":1.,"vfv":2.}
        eIndx = eIndxDic[fluxType]
        
        self.eneUnit = 'keV'
        
        fig = plt.figure(567)
        ax = fig.add_subplot(111)
            
        if logscale:
            eGrid = numpy.logspace(numpy.log10(emin),numpy.log10(emax),1000)

            ax.loglog(eGrid,eGrid**eIndx*self(eGrid),'-',**kwargs)

        else:
            eGrid = numpy.linspace(emin,emax,1000)

            ax.plot(eGrid,eGrid**eIndx*self(eGrid),'-',**kwargs)








            
         
        ax.set_xlabel("Energy [%s]"%self.eneUnit)


        
    
    def __repr__(self):
        print("Spectral model: %s" %(self.functionName))
        print("Formula:\n")
        display(Latex(self.formula))
        print("")
        print("Current parameters:\n")

        maxLen = max(map(lambda p: len(p.name) ,self.parameters.values()))

        
        table = Table(names=["Name","Value","Minimum","Maximum","Delta","Status","Unit","Prior"],dtype=["S%d"%maxLen,float,float,float,float,"S5","S6","S20"])
        for k,v in self.parameters.iteritems():
            if(v.isFree()):
                ff                   = "free"
            else:
                ff                   = "fixed"
        
            table.add_row ([v.name,v.value,v.minValue,v.maxValue,v.delta,ff,v.unit,v.prior.getName()])
        
        display(table)
    
        return ''
 
  
    @abc.abstractmethod
    def __call__(self):
        raise NotImplemented("The method __call__ has not been implemented. This is a bug")
  
    #If there is an analytical solution to the integral of the function,
    #override these methods: the analytical way is always A LOT faster
    #than numerical integration
    def photonFlux(self,e1,e2):
        return scipy.integrate.quad(self.__call__,e1,e2,epsabs=0,epsrel=1e-4)[0]
   
  
    def energyFlux(self,e1,e2):
        def eF(e):
            return e*self.__call__(e)
    
        return scipy.integrate.quad(eF,e1,e2,epsabs=0,epsrel=1e-3)[0]*keVtoErg
 
  
    #The following methods define the aritmetical operations on model, so
    #that the user can define "composite" models
    def __add__(self,b):
        return CompositeModel(self,b,operator.add)
  
    def __radd__(self,b):
        return CompositeModel(self,b,operator.add)
  
    def __sub__(self,b):
        return CompositeModel(self,b,operator.sub)
  
    def __rsub__(self,b):
        return CompositeModel(self,b,operator.sub)
  
    def __div__(self,b):
        return CompositeModel(self,b,operator.div)

    def __rdiv__(self,b):
        return CompositeModel(self,b,operator.div)
  
    def __mul__(self,b):
        return CompositeModel(self,b,operator.mul)

    def __rmul__(self,b):
        return CompositeModel(self,b,operator.mul)
  

# J. Michael
#I may have to look at how composite model is screwed up by my new scheme!!


#The following class represent the combination of two
#models with an aritmetical operation. For example,
#CompositeModel(model1,model2,operator.add) will return
#a class which will act like a (model1+model2) model.

class CompositeModel(SpectralModel):
  def __init__(self,one,two,thisOperator):
    if type(one) in (int,float):
      self.oneCall            = lambda x:float(one)
    elif issubclass(type(one),(SpectralModel,CompositeModel)):
      self.oneCall            = one.__call__
    else:
      raise NotImplementedError("Cannot execute operation with operand of type %s" %(type(one)))
    
    if type(two) in (int,float):
      self.twoCall            = lambda x:float(two)
    elif issubclass(type(two),(SpectralModel,CompositeModel)):
      self.twoCall            = two.__call__
    else:
      raise NotImplementedError("Cannot execute operation with operand of type %s" %(type(one)))
    
    self.operator             = thisOperator
    
    #Now create a parameters dictionary pointing to the
    #parameters in one and two
    self.parameters           = collections.OrderedDict()
    
    for component in [one,two]:
      if type(component) in (int,float):
        continue
      for k in component.parameters.keys():
        if(k in self.parameters.keys()):
          tokens                = k.split("_")
          if(len(tokens)==1):
            newKey              = k+"_1"
            tokens              = newKey.split("_")
          pass
          oldNum              = int(tokens[1])
          for i in range(oldNum,1000):
            newKey            = tokens[0]+"_%s" %(i)
            if(newKey not in self.parameters.keys()):
              break
            pass
          pass
        else:
          newKey                = k
        pass
        
        self.parameters[newKey] = component.parameters[k]
      pass
    pass
        
    self.formula              = "(print the components by themselves to know their formulae)"
    
    operatorNames             = {operator.add: ' + ', operator.div: ' / ', operator.mul: ' * ', operator.sub: ' - '}
    
    names                     = []
    for component in [one,two]:
      if type(component) in (int,float):
        names.insert(0,"%s" %(component))
      else:
        names.append(component.functionName)
    pass
    
    self.functionName       = "( " + operatorNames[self.operator].join(names) + " )"
    
    integrals               = []
    for component in [one,two]:
      if type(component) in (int,float):
        integrals.append(lambda e1,e2: float(component))
      else:
        integrals.append(component.integral)
      pass
    pass
    self.integral           = lambda e1,e2: integrals[0](e1,e2)+integrals[1](e1,e2)
    for k,v in self.parameters.iteritems():

        object.__setattr__(self,k,v)

   
  
  def __call__(self,x):
    return self.operator(self.oneCall(x),self.twoCall(x))
  pass

  def setup(self):
    # DO NOTHING
    pass
pass







from angsep import angsep
from disk import Disk
from ellipse import Ellipse
from gaussian import SimpleGaussian, MultiVariateGaussian
from ring import Ring, EllipticalRing
from diffusion import Diffusion



from threeML.models.spectralmodel import SpectralModel
from threeML.models.Parameter import Parameter
import math
import scipy.integrate
import scipy.special
import operator
import numpy


import collections


#test
class Gaussian(SpectralModel):

    def setup(self):
        self.functionName        = "Gaussian"
        self.formula             = r'$f(E) = A {\rm exp}\left(-\frac{(E-\mu)^2}{2\sigma^2} \right)$'
        self.parameters          = collections.OrderedDict()
        self.parameters['A']     = Parameter('A',1.,1.E-10,1.E10,0.1,fixed=False,nuisance=False,dataset=None,normalization=True)
        self.parameters['mu'] = Parameter('mu',10.,1.,1E6,0.1,fixed=False,nuisance=False,dataset=None)
        self.parameters['sigma'] = Parameter('sigma',1.,1E-2,1E1,0.1,fixed=False,nuisance=False,dataset=None)
            
        self.ncalls              = 0
        
        def integral(e1,e2):

            sigma                      = self.parameters['sigma'].value
            mu                         = self.parameters['mu'].value
            A                          = self.parameters['A'].value
            
    

            
            def f(x):

                term =  (mu - x)/(numpy.sqrt(2.)*sigma) 
                return -A* numpy.sqrt(numpy.pi/2.) * scipy.special.erf(term) * sigma
            return f(e2) - f(e1)
            
        self.integral            = integral
 
  
    def __call__(self,energy):
        self.ncalls             += 1
        sigma                      = self.parameters['sigma'].value
        mu                         = self.parameters['mu'].value
        A                          = self.parameters['A'].value
        return numpy.maximum( A * exp(-0.5 * (energy-mu)/(sigma) * (energy-mu)/(sigma)   ), 1e-30)
   
  
    def photonFlux(self,e1,e2):
        return self.integral(e1,e2)
  
    def energyFlux(self,e1,e2):
        sigma                      = self.parameters['sigma'].value
        mu                         = self.parameters['mu'].value
        A                          = self.parameters['A'].value

        
        def eF(x):
            term1 = (mu - x)/(numpy.sqrt(2.)*sigma) 
            term2 = -0.5 * (x-mu)/(sigma) * (x-mu)/(sigma)

            val = numpy.sqrt(numpy.pi/2.) * scipy.special.erf(term1) * sigma
            val += numpy.exp(term2) * sigma * sigma
             
            return -A*val

        return (eF(e2) - eF(e1))*self.keVtoErg

        




from threeML.models.tablemodel import NumpyTableModel
from threeML.models.Parameter import Parameter
import numpy
import math
import scipy.integrate
import operator

import collections

class BandTable(NumpyTableModel):
    def setup(self):

        self.SetTableFile("/Users/jburgess/Research/3ML/threeml/models/fluxModels/band.npz")


        
        self.functionName       = "Band function [Band et al. 1993]"
        self.formula            = r'''
        \[f(E) = \left\{ \begin{eqnarray}
        K \left(\frac{E}{100 \mbox{ keV}}\right)^{\alpha} & \exp{\left(\frac{-E}{E_{c}}\right)} & \mbox{ if } E < (\alpha-\beta)E_{c} \\
        K \left[ (\alpha-\beta)\frac{E_{c}}{100}\right]^{\alpha-\beta}\left(\frac{E}{100 \mbox{ keV}}\right)^{\beta} & \exp{(\beta-\alpha)} & \mbox{ if } E \ge (\alpha-\beta)E_{c}
        \end{eqnarray}
        \right.
        \]
        '''

        self.parameters         = collections.OrderedDict()
        self.parameters['alpha'] = Parameter('alpha',-1.0,-10,10,0.1,fixed=False,nuisance=False,dataset=None)
        self.parameters['beta']  = Parameter('beta',-2.0,-10,10,0.1,fixed=False,nuisance=False,dataset=None)
        self.parameters['E0']    = Parameter('E0',500,10,1e5,50,fixed=False,nuisance=False,dataset=None,unit='keV')
        self.parameters['K']     = Parameter('K',1,1e-4,1e3,0.1,fixed=False,nuisance=False,dataset=None,normalization=True)
    
    
  
  
    def __call__(self,e):
        #The input e can be either a scalar or an array
        #The following will generate a wrapper which will
        #allow to treat them in exactly the same way,
        #as they both were arrays, while keeping the output
        #in line with the input: if e is a scalar, the output
        #will be a scalar; if e is an array, the output will be an array
    
        alpha                    = self.parameters['alpha'].value
        beta                     = self.parameters['beta'].value
        E0                       = self.parameters['E0'].value
        K                        = self.parameters['K'].value
        
        return self._interpFunc((K,E0,alpha,beta,e))
    
        
 



import matplotlib.pyplot as plt
import numpy
import os
import Parameter
import collections
from threeML import fancyDisplay
from IPython.display import display, HTML

#The following color mapping has been taken from 
#http://colorbrewer2.org/?type=qualitative&scheme=Set1&n=9
#and has been selected for maximum contrast

colorPalette                  = ['#e41a1c','#377eb8','#4daf4a','#984ea3','#ff7f00','#ffff33','#a65628','#f781bf','#999999',
                                 'red','blue','cyan','green','magenta','darkgrey']



class ModelManager(object):
  def __init__(self,name,spatialModel,spectralModel,dataList):
    '''
    Implements the model for a source
    '''
    
    #Default values:
    self.emin                 = 1 #keV
    self.emax                 = 1e9 #keV
    self.nbins                = 500
        
    self.name                 = name
    self.spatialModel         = spatialModel
    self.spectralModel        = spectralModel
    self.dataList             = dataList
        
    self.parameters           = collections.OrderedDict()
    for k,v in self.spatialModel.parameters.iteritems():
      self.parameters[k]      = v
      v.setCallback(self._scheduleRecomputing)
    for k,v in self.spectralModel.parameters.iteritems():
      self.parameters[k]      = v
      v.setCallback(self._scheduleRecomputing)
    
    self.modelToBeRecomputed  = True
    
    #Link the datasets to this modelManager
    for k,dataset in self.dataList.datasets.iteritems():
      dataset.setModel(self)
      thisNuisances           = dataset.getNuisanceParameters()
      for nuisance in thisNuisances:
        nuisance.setDataset(dataset)
        self.parameters[nuisance.name] = nuisance
      pass
    pass
  pass
    
  def getParameterNames(self):
    return self.parameters.keys()
  
  def getFreeParameters(self):
    freeParameters            = collections.OrderedDict()
    
    for k,v in self.parameters.iteritems():
      if(v.isFree()):
        freeParameters[k]     = v
      pass
    pass
    
    return freeParameters    
  
  def getFreeNormalizationParameters(self):
    normParameters            = collections.OrderedDict()
    
    for k,v in self.parameters.iteritems():
      if(v.isNormalization() and v.isFree()):
        normParameters[k]     = v
      pass
    pass
    
    return normParameters 
  
  def getFreeParameterNames(self):
    freeParameters            = []
    
    for k,v in self.parameters.iteritems():
      if(v.isFree()):
        freeParameters.append(v.name)
      pass
    pass
    
    return freeParameters
  pass
    
  def getNuisanceParameterNames(self):
    return self.nuisanceParameters.keys()
  pass
  
  def __getitem__(self,paramName):
    return self.parameters[paramName]
  
  def _scheduleRecomputing(self):
    #print("Scheduling")
    self.modelToBeRecomputed  = True
  
  def setNuisanceParameterValue(self,paramName,newValue):
    self.nuisanceParameters[paramName].dataset.setNuisanceParameterValue(paramName,newValue)
    self.nuisanceParameters[paramName].setValue(newValue)
  pass
  
  def setNuisanceParameterPrior(self,paramName,newPrior):
    self.nuisanceParameters[paramName].setPrior(newPrior)
  pass
    
  def computeModel(self,emin,emax,nbins):
    #Avoid recomputing if useless
    if(self.modelToBeRecomputed==False and self.emin==emin and self.emax==emax and self.nbins==nbins):
      return
    
    self.energies             = numpy.logspace(numpy.log10(emin),numpy.log10(emax),nbins)
    self.emin                 = emin
    self.emax                 = emax
    self.nbins                = nbins
    
    #To avoid numerical issue, I fix a floor value for the model to 1e-30
    self.values               = numpy.maximum(self.spectralModel(self.energies),1e-30)
    
    #Keep the model scheduled for recomputing if I am using a non-standard energy
    #bin
    self.modelToBeRecomputed  = False
  pass
    
  def plot(self,kind="counts",emin=1.0,emax=1e12,nbins=1000,**kwargs):
    #Compute the model with the current parameter values
    self.modelToBeRecomputed  = True
    self.computeModel(emin,emax,nbins)
    self.modelToBeRecomputed  = True
    sub                       = None
    
    #By default plot all datasets
    datasets                  = self.dataList.datasets.keys()
    for k,v in kwargs.iteritems():
      if(k.lower()=="subplot"):
        sub                   = v
      elif(k.lower()=="datasets"):
        try:
          datasets              = list(v)
        except:
          raise RuntimeError("You have to provide a list as argument for the 'datasets' keyword.")
      pass
    pass
    
    #plot it
    legend                    = False
    if(kind.lower()=="flux"):
      x                       = [self.energies]
      y                       = [self.values]
      xerr                    = [[]]
      yerr                    = [[]]
      labels                  = ['']
      alphas                  = [1]
      colors                  = [colorPalette[0]]
      xlabel                  = "keV"
      ylabel                  = r"Photons cm$^{-2}$ s$^{-1}$ keV$^{-1}$"
      styles                  = ['-']
    elif(kind.lower()=="nufnu"):
      x                       = [self.energies]
      keVtoErg                = 1.60217657e-9
      y                       = [self.values*numpy.power(numpy.array(self.energies),2)*keVtoErg]
      xerr                    = [[]]
      yerr                    = [[]]
      labels                  = ['']
      alphas                  = [1]
      colors                  = [colorPalette[0]]
      xlabel                  = "keV"
      ylabel                  = r"$\nu$ F$_{\nu}$ [erg cm$^{-2}$ s$^{-1}$]"
      styles                  = ['-']
    elif(kind.lower()=="counts"):
      #Counts plot
      legend                  = True
      x                       = []
      xerr                    = []
      y                       = []
      yerr                    = []
      styles                  = []
      labels                  = []
      alphas                  = []
      colors                  = []
      for i,(k,dataset) in enumerate(self.dataList.datasets.iteritems()):
        if(not k in datasets):
          continue
        labels.append(k)
        folded,e1s,e2s,counts = dataset.getModelAndData()
        xmean                 = (e1s+e2s)/2.0
        de                    = (e2s-e1s)/2.0
        #Folded counts
        x.append(xmean)
        xerr.append(de)
        y.append(counts)
        yerr.append(numpy.sqrt(counts))
        styles.append(',')
        alphas.append(1)
        colors.append(colorPalette[i])
        #Model
        x.append(xmean)
        xerr.append([])
        y.append(folded)
        yerr.append([])
        styles.append('step')
        labels.append('')
        alphas.append(0.75)
        colors.append(colorPalette[i])
      pass
      
      xlabel                  = "keV"
      ylabel                  = r"Counts"
    else:
      raise RuntimeError("Unknown plot kind") 
    
    if(sub==None):
      fig,sub                   = plt.subplots(1,1)
    else:
      fig                       = sub.get_figure()
    pass
    
    for i in range(len(x)):
      args                    = {}
      if(legend and labels[i]!=''):
        args['label']         = labels[i]
      pass
      args['color']           = colors[i]
      args['alpha']           = alphas[i]
      
      if(len(xerr[i])==0 and len(yerr[i])==0):
        if(styles[i]=="step"):
          sub.step(x[i],y[i],where='mid',**args)
        else:
          sub.plot(x[i],y[i],styles[i],**args)
      else:
        sub.errorbar(x[i],y[i],xerr=xerr[i],yerr=yerr[i],fmt=styles[i],capsize=0,**args)
      pass
    pass
    
    #Optimize plot ranges
    
    
    sub.set_xscale("log",nonposx='clip')
    sub.set_yscale("log",nonposx='clip')
    sub.set_xlabel(xlabel)
    sub.set_ylabel(ylabel)
    sub.axis('tight')
    lims                      = plt.xlim()
    sub.set_xlim([lims[0]/2,lims[1]*2])
    lims                      = plt.ylim()
    sub.set_ylim([lims[0]/2,lims[1]*2])
    
    if(legend):
      sub.legend(loc=0,numpoints=1,ncol=2,fontsize='small')
    return fig
  pass
  
  def plotWithChain(self,parNames,samples,**kwargs):
    
    #Default values
    kind                      = 'flux'
    ymin                      = 1e-18
    emin                      = 1.0
    emax                      = 1e12
    nbins                     = 1000
    percentiles               = [16.0,50.0,84.0]
    subplot                   = None
    for k,v in kwargs.iteritems():
      if(k.lower()=="kind"):
        kind                  = v.lower()
      elif(k.lower()=="ymin"):
        ymin                  = float(v)
      elif(k.lower()=="emin"):
        emin                  = float(v)
      elif(k.lower()=="emax"):
        emax                  = float(v)
      elif(k.lower()=="nbins"):
        nbins                 = int(v)
      elif(k.lower()=="subplot"):
        subplot               = v
      elif(k.lower()=="percentiles"):
        percentiles           = list(v)
      elif(k.lower()=="energies"):
        #Use it later on 
        pass
      else:
        raise RuntimeError("Unknown keyword %s" %(k))
      pass
    pass
        
    #Compute the model with the current parameter values
    energies                  = numpy.logspace(numpy.log10(emin),numpy.log10(emax),nbins)
    if("energies" in map(lambda x:x.lower(),kwargs.keys())):
      energies                = kwargs["energies"]
    pass
    
    #Compute the quantiles
    #Compute the models for all the set of parameters
    
    modelValues               = numpy.zeros(shape=[energies.shape[0],len(samples)])
    for i,sample in enumerate(samples):
      map(lambda (key,val):self[key].setValue(val),
                                    zip(parNames,sample))      
      modelValues[:,i]        = self.spectralModel(energies)
    pass
    
    #Now for each energy get the quantiles
    median                    = numpy.zeros(energies.shape[0])
    lowQuantile               = numpy.zeros(energies.shape[0])
    hiQuantile                = numpy.zeros(energies.shape[0])
    for i in range(energies.shape[0]):
      l,m,h                   = numpy.percentile(modelValues[i,:],percentiles)
      median[i]               = m
      lowQuantile[i]          = l
      hiQuantile[i]           = h
    pass
    
    self.modelValues          = modelValues
    
    #Plot everything
    if(kind.lower()=="flux"):
      x                       = energies
      conv                    = 1.0
      xlabel                  = "keV"
      ylabel                  = r"Photons cm$^{-2}$ s$^{-1}$ keV$^{-1}$"
    elif(kind.lower()=="nufnu"):
      x                       = energies
      keVtoErg                = 1.60217657e-9
      conv                    = numpy.power(numpy.array(x)*keVtoErg,2)
      xlabel                  = "keV"
      ylabel                  = r"$\nu$ F$_{\nu}$ [erg cm$^{-2}$ s$^{-1}$]"
    pass    
    
    if(subplot==None):
      fig,sub                   = plt.subplots(1,1)
    else:
      sub                       = subplot
      fig                       = sub.get_figure()
    pass
    
    sub.fill_between(energies,lowQuantile*conv,hiQuantile*conv,facecolor='green', alpha=0.4,linewidth=0.0)
    sub.plot(x,median*conv,'--')
    sub.loglog()
    sub.set_xlabel(xlabel)
    sub.set_ylabel(ylabel)
    sub.set_ylim([ymin,max(hiQuantile*conv)*1.1])
    
    self.median               = median
    self.lowQuantile          = lowQuantile
    self.hiQuantile           = hiQuantile
    return fig    
  pass
  
  def __repr__(self):
    print self.spatialModel
    print self.spectralModel
    
    toPrint                   = []
    for k,v in self.parameters.iteritems():
      if(v.isNuisance()):
        toPrint.append(v)
      pass
    pass
    
    print("\nNuisance parameters:\n")
    table                    = fancyDisplay.HtmlTable(7)
    table.addHeadings("Name","Value","Minimum","Maximum","Delta","Dataset","Unit")
    for v in toPrint:
      table.addRow(v.name,v.value,v.minValue,v.maxValue,v.delta,v.dataset.getName(),v.unit)
    pass
    display(HTML(table.__repr__()))

    return ''
  pass
  
  def printParamValues(self,output=False):
    string                    = []
    for k,v in self.parameters.iteritems():
      string.append("%s = %s" %(k,v.value))
    pass
    
    if(output):
      print(",".join(string))
    else:
      return ",".join(string)
  pass
  
  def writeToFile(self,filename,emin=None,emax=None,nbins=None):
    '''
    Write the spectral model to a file compatible with gtlike
    '''
    #Compute the model with the current parameter values
    self.computeModel(emin,emax,nbins)
    
    #Write it to file
    with open(filename,"w+") as f:
      modelString             = "#%s\n" % (self.spectralModel.functionName)
      for k,v in self.parameters.iteritems():
        modelString          += "#%s\n" % v.__repr__()
      pass
      for e,v in zip(self.energies,self.values):
        #Gtlike needs the energy in MeV and the flux in ph/cm2/s/MeV
        modelString          += "%.20g %.20g\n" %(e/1000.0,v*1000.0)
      f.write(modelString)
  pass  
pass



from PointSource import PointSource
from ExtendedSource import ExtendedSource
import collections
import numpy

class MyCollection(object):
  '''
  An ordered collections of key,value pairs which can be accessed
  by name or by numeral id
  '''
  def __init__(self):
    self.dict                 = collections.OrderedDict()
  
  def __getitem__(self,id_or_name):
    
    if(isinstance(id_or_name,int)):
      
      #by ID
      return self.dict.values()[id_or_name]
    
    elif(isinstance(id_or_name,str)):
      
      #By name
      return self.dict[id_or_name]
    
    else:
      raise RuntimeError("%s must be either an int or a str" %(id_or_name))
   
  def __setitem__(self,key,val):
     if(isinstance(key,int)):
       raise RuntimeError("Cannot use integers as key")
     
     self.dict[key]           = val
  
  def __len__(self):
     return len(self.dict.keys())
  
class LikelihoodModel(object):
  
  def __init__(self,*sources):
    
    if(len(sources)==0):
      raise RuntimeError("You have to provide at least one source")
    
    #Store sources in ordered dictionaries, so that we can
    #access them by name or by id (ordinal number)
    
    self.pointSources         = MyCollection()
    self.extendedSources      = MyCollection()
    
    #Make also a dictionary of all parameters
    self.parameters         = collections.OrderedDict()
    
    #Loop through the provided sources
    for source in sources:
      
      if(isinstance(source,ExtendedSource)):
        
        #Current source is an Extended Source
        
        self.extendedSources[source.name] = source
      
      elif(isinstance(source,PointSource)):
        
        #Current source is a point source
        
        self.pointSources[source.name]    = source
      
      else:
      
        #User error, source is not a source (!)
        
        raise RuntimeError("One of the argument provided is not a "+
                           "point source nor an extended source")
      
      pass
      
      #Now get the parameters and add them to the
      #internal dictionary
      
      self.parameters[source.name] = source.getAllParameters()
      
    pass
  pass
  
  def getFreeParameters(self):
    
    freeParameters            = collections.OrderedDict()
    
    for srcName,srcParameters in self.parameters.iteritems():
      
      for parName,par in srcParameters.iteritems():
        
        if(par.isFree() and (not par.isNuisance())):
          
          freeParameters[(srcName,parName)]     = par
    
    return freeParameters
  
  def getNumberOfPointSources(self):
    
    return len(self.pointSources) 
  
  def getPointSourcePosition(self,id_or_name):
    '''
    Return the position of the point source. You can use either the name
    or the ID of the source
    '''
    return self.pointSources[id_or_name].getPosition()
  pass
      
  def getPointSourceFluxes(self,id_or_name, energies):
    return self.pointSources[id_or_name].getFlux( numpy.array(energies) )
  
  def getPointSourceName(self,id_or_name):
    return self.pointSources[id_or_name].name
  
  ##############################################
  
  def getNumberOfExtendedSources(self):
    return len(self.extendedSources)

  def getExtendedSourceFluxes(self, srcid, j2000_ra, j2000_dec, energies):
    
    return self.extendedSources[srcid].getBrightness( j2000_ra, j2000_dec, energies)
  
  def getExtendedSourceName(self,id_or_name):
    
    return self.extendedSources[id_or_name].name
  
  def getExtendedSourceBoundaries( self, id_or_name ):
    
    ramin, ramax, decmin, decmax = self.extendedSources[id_or_name].getBoundaries()
    
    return numpy.array( [ramin, ramax, decmin, decmax]  )
  
  def isInsideAnyExtendedSource( j2000_ra,  j2000_dec):
    
    return true
  
    
